@@ -1,4 +1,4 @@
-/* Copyright (c) 2014, 2024, Oracle and/or its affiliates.
+/* Copyright (c) 2014, 2025, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2.0,
@@ -84,8 +84,11 @@ int Certifier_broadcast_thread::initialize() {
   if ((mysql_thread_create(key_GR_THD_cert_broadcast, &broadcast_pthd,
                            get_connection_attrib(), launch_broadcast_thread,
                            (void *)this))) {
-    mysql_mutex_unlock(&broadcast_run_lock); /* purecov: inspected */
-    return 1;                                /* purecov: inspected */
+    /* purecov: begin inspected */
+    mysql_mutex_unlock(&broadcast_run_lock);
+    LogPluginErr(ERROR_LEVEL, ER_GRP_RPL_CERT_BROADCAST_THREAD_CREATE_FAILED);
+    return 1;
+    /* purecov: end */
   }
   broadcast_thd_state.set_created();
 
@@ -98,13 +101,13 @@ int Certifier_broadcast_thread::initialize() {
   return 0;
 }
 
-int Certifier_broadcast_thread::terminate() {
+void Certifier_broadcast_thread::terminate() {
   DBUG_TRACE;
 
   mysql_mutex_lock(&broadcast_run_lock);
   if (broadcast_thd_state.is_thread_dead()) {
     mysql_mutex_unlock(&broadcast_run_lock);
-    return 0;
+    return;
   }
 
   aborted = true;
@@ -122,8 +125,6 @@ int Certifier_broadcast_thread::terminate() {
     mysql_cond_wait(&broadcast_run_cond, &broadcast_run_lock);
   }
   mysql_mutex_unlock(&broadcast_run_lock);
-
-  return 0;
 }
 
 void Certifier_broadcast_thread::dispatcher() {
@@ -143,6 +144,8 @@ void Certifier_broadcast_thread::dispatcher() {
   mysql_cond_broadcast(&broadcast_run_cond);
   mysql_mutex_unlock(&broadcast_run_lock);
 
+  LogPluginErr(SYSTEM_LEVEL, ER_GRP_RPL_CERT_BROADCAST_THREAD_STARTED);
+
   while (!aborted) {
     // Broadcast Transaction identifiers every 30 seconds
     if (broadcast_counter % 30 == 0) {
@@ -196,6 +199,8 @@ void Certifier_broadcast_thread::dispatcher() {
   mysql_cond_broadcast(&broadcast_run_cond);
   mysql_mutex_unlock(&broadcast_run_lock);
 
+  LogPluginErr(SYSTEM_LEVEL, ER_GRP_RPL_CERT_BROADCAST_THREAD_STOPPED);
+
   my_thread_exit(nullptr);
 }
 
@@ -306,6 +311,10 @@ Certifier::Certifier()
 Certifier::~Certifier() {
   mysql_mutex_lock(&LOCK_certification_info);
   initialized = false;
+
+  broadcast_thread->terminate();
+  delete broadcast_thread;
+
   clear_certification_info();
   delete certification_info_sid_map;
 
@@ -316,7 +325,6 @@ Certifier::~Certifier() {
   delete group_gtid_extracted;
   delete group_gtid_sid_map;
   mysql_mutex_unlock(&LOCK_certification_info);
-  delete broadcast_thread;
 
   mysql_mutex_lock(&LOCK_members);
   clear_members();
@@ -640,15 +648,6 @@ end:
   return error;
 }
 
-int Certifier::terminate() {
-  DBUG_TRACE;
-  int error = 0;
-
-  if (is_initialized()) error = broadcast_thread->terminate();
-
-  return error;
-}
-
 void Certifier::update_parallel_applier_indexes(
     bool update_parallel_applier_last_committed_global,
     bool increment_parallel_applier_sequence_number) {
@@ -926,6 +925,8 @@ rpl_gno Certifier::certify(Gtid_set *snapshot_version,
     Update parallel applier indexes.
   */
   if (!local_transaction) {
+    bool is_empty_transaction = false;
+
     /*
       'CREATE TABLE ... AS SELECT' is considered a DML, though in reality it
       is DDL + DML, which write-sets do not capture all dependencies.
@@ -938,7 +939,18 @@ rpl_gno Certifier::certify(Gtid_set *snapshot_version,
       update_parallel_applier_last_committed_global = true;
     }
 
-    if (!has_write_set || update_parallel_applier_last_committed_global) {
+    /*
+      Empty transactions, despite not having write-set, can be
+      applied in parallel with any other transaction.
+      Empty transactions are assigned `last_committed = -1` by GR
+      before send.
+    */
+    else if (!has_write_set && -1 == gle->last_committed) {
+      is_empty_transaction = true;
+    }
+
+    if (!is_empty_transaction &&
+        (!has_write_set || update_parallel_applier_last_committed_global)) {
       /*
         DDL does not have write-set, so we need to ensure that it
         is applied without any other transaction in parallel.
@@ -953,7 +965,9 @@ rpl_gno Certifier::certify(Gtid_set *snapshot_version,
     assert(gle->last_committed < gle->sequence_number);
 
     update_parallel_applier_indexes(
-        !has_write_set || update_parallel_applier_last_committed_global, true);
+        (!is_empty_transaction &&
+         (!has_write_set || update_parallel_applier_last_committed_global)),
+        true);
   }
 
 end:
